# -*- coding: utf-8 -*-

import subprocess
import os.path
import re


class ProbeFailedException(Exception):
  """Exception that is raised if ffprobe failed"""
  def __init__(self, retcode):
    super(ProbeFailedException, self).__init__(u"Probe of video failed with return code %d" % retcode)


class ParseFailedException(Exception):
  """Exception that is raised if parse of ffprobe output failed"""
  def __init__(self, output):
    super(ParseFailedException, self).__init__(u"Parse of ffprobe output video failed:\n%s" % output)


class InvalidTimecodeException(Exception):
  """Exception that is raised if the timecode is out of bounds"""
  def __init__(self, timecode):
    super(InvalidTimecodeException, self).__init__("Invalid timecode: %d" % timecode)


class InvalidRegionException(Exception):
  """Exception that is raised if a region's start is >= its end"""
  def __init__(self, start, end):
    super(InvalidRegionException, self).__init__("Region's end is less or equal to start: %d-%d" % (start, end))


class RegionStraddlesSegmentsException(Exception):
  """Exception that is raised if a region straddles segment boundaries"""
  def __init__(self, start, end):
    super(RegionStraddlesSegmentsException, self).__init__("Region straddles segments: %d-%d" % (start, end))
    

class Video(object):
  """Abstracts a video, even if it is in multiple segments"""
  filelist_pattern = re.compile(r"""%[0-9]+d""")
  
  def __init__(self, video, ffprobe='/usr/bin/ffprobe'):
    """Constructs a new video on one or more underlying files.
       video is either a single filename, or a list of filenames
       that specify consecutive segments of the video.
    """
    if hasattr(video, "__iter__"):
      self.videolist = video
    else:
      self.videolist = [video] 
    self._check_videos_exist()
    self.durations = None
    self.cumulative_duration = None
    self.sizes = None
    self.cumulative_size = None
    self.ffprobe = ffprobe
    self.use_show_format = self._support_show_format()
  
  def get_duration(self):
    """Returns the duration of the video in milliseconds"""
    # Note: Internally the durations are stored in microseconds for greater accuracy
    self._get_metadata()
    return int(round(self.cumulative_duration / 1000.0))
  
  def get_size(self):
    """Returns the size of the video in bytes"""
    self._get_metadata()
    return self.cumulative_size
  
  def get_fps(self):
    """Returns the frame rate of the video"""
    self._get_metadata()
    return self.fps
  
  def get_dimensions(self):
    """Returns the (width, height) of the video"""
    self._get_metadata()
    return self.dim
  
  def get_segment_and_offset(self, timecode, clip=False):
    """Returns the video segment to which the timecode in milliseconds belongs,
      and the offset relative to the start of the segment"""
    duration = self.get_duration()
    if timecode < 0:
      if not clip:
        raise InvalidTimecodeException(timecode)
      else:
        timecode = 0
    elif timecode > duration:
      if not clip:
        raise InvalidTimecodeException(timecode)
      else:
        timecode = duration
    
    timecode_in_us = timecode * 1000
    start_timecode_in_us = 0
    for video in self.videolist:
      seg_duration = self.durations[video]
      video_end_timecode_in_us = start_timecode_in_us + seg_duration
      if timecode_in_us >= start_timecode_in_us and timecode_in_us < video_end_timecode_in_us:
        return (video, timecode - int(round(start_timecode_in_us / 1000.0)))
      start_timecode_in_us = video_end_timecode_in_us
    # timecode equals duration
    return(self.videolist[-1], int(round(self.durations[self.videolist[-1]] / 1000.0)))
  
  def get_region(self, start, end, clip=False):
    """Returns the video segment and timecodes relative to the start of
       the video that covers a region from start to end, inclusive. start
       and end may not straddle multiple videos, with one exception:
       if end touches exactly a segment boundary, it is accepted.
    """
    if end <= start:
      raise InvalidRegionException(start, end)
    (start_video, start_timecode) = self.get_segment_and_offset(start, clip)
    (end_video, end_timecode) = self.get_segment_and_offset(end, clip)
    if start_video == end_video:
      return (start_video, start_timecode, end_timecode)
    else:
      # if two segments are adjacent, and the end is just past the end
      # of the first segment; i.e. the start of the second segment, this
      # is acceptable, too
      if self.videolist.index(start_video) == self.videolist.index(end_video) - 1 and \
        end_timecode == 0:
        return (start_video, start_timecode, int(round(self.durations[start_video] / 1000.0)))
      else:
        raise RegionStraddlesSegmentsException(start, end)
  
  def _check_videos_exist(self):
    # checks whether all videos exist
    for video in self.videolist:
      if self.filelist_pattern.search(video):
        # video is a filelist
        video = os.path.dirname(video)
      if not os.path.exists(video):
        raise IOError(2, "No such file or directory: '%s'" % video)

  def _get_metadata(self):
    # Retrieves the metadata of the videos
    if self.durations is None:
      self.durations = dict()
      self.sizes = dict()
      for video in self.videolist:
        format_switch = '-show_format' if self.use_show_format else '-show_files'
        ffprobe_process = subprocess.Popen([self.ffprobe, format_switch, '-show_streams', video],
                                           stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE, close_fds=True)
        (stdout, stderr) = ffprobe_process.communicate()
        if ffprobe_process.returncode != 0:
          raise ProbeFailedException(ffprobe_process.returncode)
        (duration, size, fps, dim) = self._parse_metadata(stdout)
        duration = int(round(duration * 1000000))
        self.durations[video] = duration
        self.sizes[video] = size
        self.fps = fps
        self.dim = dim
    self.cumulative_duration = sum(self.durations.itervalues())
    self.cumulative_size = sum(self.sizes.itervalues())
  
  def _parse_metadata(self, text):
    # Parses the meta of the stream
    duration = None
    size = None
    fps = None
    width = None
    height = None
    in_video = True
    for line in text.split("\n"):
      if line.startswith("[/STREAM]"):
        in_video = False
      if line.startswith("codec_type=video"):
        in_video = True
      if in_video:
        if line.startswith("duration="):
          duration = float(line[len("duration="):])
        if self.use_show_format:
          if line.startswith("r_frame_rate="):
            info = line[len("r_frame_rate="):]
            try:
              fps = float(info[:info.rindex("/")]) / float(info[info.rindex("/") + 1:])
            except ZeroDivisionError:
              fps = -1
        else:
          if line.startswith("r_frame_rate_num="):
            fps = float(line[len("r_frame_rate_num="):])
          if line.startswith("r_frame_rate_den="):
            fps /= float(line[len("r_frame_rate_den="):])
        if line.startswith("width="):
          width = int(line[len("width="):])
        if line.startswith("height="):
          height = int(line[len("height="):])
      if line.startswith("size="):
        size = int(float(line[len("size="):]))
      if fps is not None and duration is not None \
        and size is not None:
        return (duration, size, fps, (width, height))
    # we should never get here
    raise ParseFailedException(text)
  
  def _support_show_format(self):
    # Checks whether ffprobe supports -show_format (which is the
    # case when it is integrated with ffmpeg.
    ffprobe_process = subprocess.Popen([self.ffprobe, '-show_format'],
                                           stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE, close_fds=True)
    (stdout, stderr) = ffprobe_process.communicate()
    return stderr.find("unrecognized option '-show_format'") < 0
