package edu.bu.signstream.common.util;

import edu.bu.signstream.common.util.vo.CustomPanelResizeEvent;
import edu.bu.signstream.common.util.vo.CustomPanelResizeListener;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;

/**
 * This class provides a custom implementation of a split pane. It allows for the resizing of panels in a split pane, along with some other features.
 * The biggest feature is having direct access to panels within the splitpane in a much easier way, and having n number of panels in 1 split pane rather than the
 * Java Swing default of 2 panels and then nesting split panes to have multiple panels within a split pane.
 */
public class CustomPanelizedSplitPane extends JPanel {
    private ArrayList<Double> panelRatios;
    private ArrayList<JPanel> panels;
    private ArrayList<JPanel> dividers;
    public static final int DIVIDER_SIZE = 5; // Minimum height of divider
    public static final int MIN_PANEL_SIZE = 50; // The minimum size of a panel in pixels.
    private final ArrayList<CustomPanelResizeListener> resizeListeners = new ArrayList<>();

    public CustomPanelizedSplitPane(ArrayList<JPanel> panels, ArrayList<Double> panelRatios) {
        // Create the outer container panel
        super.setLayout(null);
        this.panels = panels;
        this.panelRatios = panelRatios;
        this.normalizeRatios();

        this.dividers = new ArrayList<>();

        for (int i = 0; i < panels.size() - 1; i++) {
            this.dividers.add(this.createDivider(this, i));
        }

        for (JPanel panel : panels) this.add(panel);
        for (JPanel divider : this.dividers) this.add(divider);

        super.addComponentListener(new ComponentAdapter() {
            @Override
            public void componentResized(ComponentEvent e) {
                // CustomPanelizedSplitPane.this.revalidateRatios();
                CustomPanelizedSplitPane.this.resizeComponents();
            }
        });

        this.resizeComponents();
    }

    /**
     * Converts ratios greater than 1.0 to a decimal value.
     * 50% = 0.5
     */
    private void normalizeRatios() {
        for (int x = 0; x < this.panelRatios.size(); x++) {
            if (this.panelRatios.get(x) > 1.0) {
                this.panelRatios.set(x, this.panelRatios.get(x) / 100.0);
            }
        }
    }

    /**
     * When the outer panel is resized, this method will check if any of the panels are under the minimum height ratio, and if so, it will adjust the ratios so that the minimum height ratio is the minimum height ratio
     * The reason it is not used is that when the outer panel is expanded, we lose the original panel ratios and original proportions...
     * If we have a 50, 25, 25 ratio, then as we revalidate it when the panel gets smaller, the ratios also change accordingly to make sure
     * each panel maintains minimum size, but when we expand the panel again, we lose the original ratios and proportions, so the panels won't be
     * the same size they were before. So, for now, because the logic is tricky to implement when expanding the outer panel back out,
     * we are not using this method.
     */
    private void revalidateRatios() {
        int height = super.getHeight();
        double minimumHeightRatio = MIN_PANEL_SIZE / (double) height;
        // For each panelRatio, if it is under the minimum height ratio, then we need to adjust the ratios so that the minimum height ratio is the minimum height ratio
        ArrayList<Double> panelRatios = new ArrayList<>(this.panelRatios);
        boolean canRevalidate = false;
        // First check if any panels are above the minimum height ratio
        for (Double ratio : this.panelRatios) {
            if (ratio > minimumHeightRatio) {
                canRevalidate = true;
                break;
            }
        }

        if (!canRevalidate) {
            return;
        }

        double totalAdded = 0.0;
        for (int x = 0; x < this.panelRatios.size(); x++) {
            if (this.panelRatios.get(x) < minimumHeightRatio) {
                double difference = minimumHeightRatio - this.panelRatios.get(x);
                this.panelRatios.set(x, minimumHeightRatio);
                totalAdded += difference;

                System.out.println("Difference: " + difference + " | Minimum Height Ratio: " + minimumHeightRatio + " | Panel Ratio: " + this.panelRatios.get(x) + " | Outer Panel Height: " + height);
            }
        }

        for (int x = 0; x < this.panelRatios.size(); x++) {
            if (this.panelRatios.get(x) > minimumHeightRatio && totalAdded > 0.0) {
                // Check if we can subtract all of the totalAdded from this panel ratio, otherwise we need to subtract the difference between the panel ratio and the minimum height ratio
                double difference = Math.min(totalAdded, this.panelRatios.get(x) - minimumHeightRatio);
                this.panelRatios.set(x, this.panelRatios.get(x) - difference);
                totalAdded -= difference;
            }
        }

        if (totalAdded > 0.0) {
            // We still have some left over, so restore to the original panel ratios
            this.panelRatios = panelRatios;
        }
    }


    /**
     * @param outerPanel The outer panel that contains the panels and dividers
     * @param index     The index of the divider relative to the other dividers
     * @return
     */
    private JPanel createDivider(JPanel outerPanel, int index) {
        JPanel divider = new JPanel();
        divider.setBackground(new Color(211, 211, 211)); // Light gray
        divider.setCursor(new Cursor(Cursor.S_RESIZE_CURSOR));

        divider.addMouseMotionListener(new MouseMotionAdapter() {
            /**
             * This method will resize the panels above and below the divider based on the mouse drag, pushing panels up or down as appropriate once a specific panel reaches MIN_PANEL_SIZE
             * @param e the event to be processed
             */
            @Override
            public void mouseDragged(MouseEvent e) {
                int mouseY = e.getYOnScreen() + DIVIDER_SIZE;
                int frameY = outerPanel.getLocationOnScreen().y;

                double accumulatedHeight = 0; // The accumulated height of the panels above the divider
                for (int i = 0; i < index; i++) {
                    accumulatedHeight += (CustomPanelizedSplitPane.this.panelRatios.get(i) * outerPanel.getHeight()) + DIVIDER_SIZE; // Do we need to account for the dividers here??
                }

                double newPanelRatio = ((mouseY - frameY) - accumulatedHeight) / outerPanel.getHeight(); // How large the new panel should be relative to the outer panel and mouse location of the drag

                // Calculate the difference from the original ratio for the resized panel
                double difference = newPanelRatio - CustomPanelizedSplitPane.this.panelRatios.get(index); // The amount of change in the panel ratio for the top panel (above the divider)

                if (CustomPanelizedSplitPane.this.isResizeValid(outerPanel, index, difference)) {
                    CustomPanelizedSplitPane.this.panelRatios.set(index, CustomPanelizedSplitPane.this.panelRatios.get(index) + difference); // The panel above the divider

                    // Distribute the size difference to the panels below the current panel
                    double remainingDiff = difference;
                    for (int i = index + 1; i < CustomPanelizedSplitPane.this.panelRatios.size() && remainingDiff != 0; i++) {
                        double panelHeight = CustomPanelizedSplitPane.this.panels.get(i).getHeight(); // For any of the panels, the height is - the total divider sizes because they are overlayed on the panel
                        double potentialChange = Math.min(remainingDiff, (outerPanel.getHeight() - panelHeight / outerPanel.getHeight()));

                        if (panelHeight - potentialChange * outerPanel.getHeight() < MIN_PANEL_SIZE) { // If the panel will be less than the minimum size, then we need to adjust the change and only subtract what we can
                            potentialChange = (panelHeight - MIN_PANEL_SIZE) / outerPanel.getHeight();
                        }

                        CustomPanelizedSplitPane.this.panelRatios.set(i, CustomPanelizedSplitPane.this.panelRatios.get(i) - potentialChange);
                        remainingDiff -= potentialChange;
                    }
                } else {
                    this.pushPanels(outerPanel, difference); // We can't do the resize without violating the minimum size, so we need to push the panels above and below the divider
                }

                CustomPanelizedSplitPane.this.resizeComponents();
                CustomPanelizedSplitPane.this.fireResizeEvent();
            }

            /**
             * This method will push the panels above and below the divider to make room for the panel that is being resized, depending on the direction of the resize.
             * @param outerPanel The outer panel that contains the panels and dividers
             * @param difference The amount of change in the panel ratio for the top panel (above the divider)
             */
            private void pushPanels(JPanel outerPanel, double difference) {
                double accumulatedDifference = difference;
                // When pushing up, the panel above will get bigger (potential difference is negative)
                // When pushing down, the panel above will get smaller (potential difference is positive)
                if (difference < 0) {  // Pushing up
                    for (int i = index - 1; i >= 0 && accumulatedDifference < 0; i--) {
                        double potentialChange = Math.min(-accumulatedDifference, (double) (CustomPanelizedSplitPane.this.panels.get(i).getHeight() - MIN_PANEL_SIZE) / outerPanel.getHeight());

                        if (i == 0 && CustomPanelizedSplitPane.this.panels.get(i).getHeight() - potentialChange * outerPanel.getHeight() <= MIN_PANEL_SIZE) {
                            return; // topmost panel reached minimum size, can't push up anymore
                        }

                        CustomPanelizedSplitPane.this.panelRatios.set(i, CustomPanelizedSplitPane.this.panelRatios.get(i) - potentialChange); // Set the panel above this divider to make space for panel below it we are pushing up
                        CustomPanelizedSplitPane.this.panelRatios.set(index + 1, CustomPanelizedSplitPane.this.panelRatios.get(index + 1) + potentialChange); // Set the panel below this divider to take up the space we are pushing up
                        accumulatedDifference += potentialChange;
                    }

                    CustomPanelizedSplitPane.this.resizeComponents();
                } else if (difference > 0) {  // Pushing down
                    for (int i = index + 1; i < CustomPanelizedSplitPane.this.panelRatios.size() && accumulatedDifference > 0; i++) {
                        double potentialChange = Math.min(accumulatedDifference, (double) (CustomPanelizedSplitPane.this.panels.get(i).getHeight() - MIN_PANEL_SIZE) / outerPanel.getHeight());

                        if (i == CustomPanelizedSplitPane.this.panelRatios.size() - 1 && outerPanel.getHeight() - CustomPanelizedSplitPane.this.dividers.get(CustomPanelizedSplitPane.this.dividers.size() - 1).getY() <= MIN_PANEL_SIZE) {
                            return; // bottom-most panel reached minimum size, can't push down anymore
                        }

                        CustomPanelizedSplitPane.this.panelRatios.set(i, CustomPanelizedSplitPane.this.panelRatios.get(i) - potentialChange); // Set the panel below this divider to make space for panel above it we are pushing down
                        CustomPanelizedSplitPane.this.panelRatios.set(index, CustomPanelizedSplitPane.this.panelRatios.get(index) + potentialChange); // Set the panel above this divider to take up the space we are pushing down
                        accumulatedDifference -= potentialChange;
                    }

                    CustomPanelizedSplitPane.this.resizeComponents();
                }
            }
        });

        return divider;
    }

    /**
     * @param outerPanel  The outer panel that contains the panels and dividers
     * @param dividerIndex The index of the divider relative to the other dividers
     * @param difference The amount of change in the panel ratio for the top panel (above the divider)
     * @return Whether the resize is valid — whether as a result of the resize, the panel above and below will still be >= the minimum size of the panel
     */
    private boolean isResizeValid(JPanel outerPanel, int dividerIndex, double difference) {
        double differenceSize = difference * outerPanel.getHeight();

        // Check if the panel above the divider will still be at least the minimum size after resizing with differenceSize
        boolean cond1 = (this.panels.get(dividerIndex).getHeight() + differenceSize >= MIN_PANEL_SIZE);
        // Check if the panel below the divider will still be at least the minimum size after resizing with differenceSize
        boolean cond2 = (this.panels.get(dividerIndex + 1).getHeight() - differenceSize >= MIN_PANEL_SIZE);

        return cond1 && cond2;
    }


    /**
     * Resizes the components based on the current panel ratios.
     */
    public void resizeComponents() {
        if (this.panels.isEmpty()) {
            return;
        }

        int frameHeight = this.getHeight();
        int frameWidth = this.getWidth();
        int minimumHeight = (this.panels.size() * MIN_PANEL_SIZE) + ((this.panels.size() - 1) * DIVIDER_SIZE);

        frameHeight = this.getHeight() < minimumHeight ? minimumHeight : frameHeight;
        super.setPreferredSize(new Dimension(frameWidth, frameHeight));
        super.setSize(super.getPreferredSize());

        int extraSpacePerPanel = DIVIDER_SIZE / this.panels.size();
        int remainder = DIVIDER_SIZE % this.panels.size();

        int lastPos = 0;
        for (int i = 0; i < this.panels.size(); i++) {
            int panelHeight = (int) (this.panelRatios.get(i) * super.getHeight());

            if (i < this.dividers.size()) {
                panelHeight -= DIVIDER_SIZE;
            } else {
                panelHeight += remainder; // Add remainder to the last panel
            }

            panelHeight += extraSpacePerPanel; // Distribute the missing divider size

            this.panels.get(i).setBounds(0, lastPos, frameWidth, panelHeight);
            lastPos += panelHeight;

            if (i < this.dividers.size()) {
                this.dividers.get(i).setBounds(0, lastPos, frameWidth, DIVIDER_SIZE);
                lastPos += DIVIDER_SIZE;
            }
        }

        for (JPanel panel : this.panels) {
            panel.revalidate();
            panel.repaint();
        }

        super.revalidate();
        super.repaint();
    }

    /**
     * @param panels     The panels to set
     * @param panelRatios The ratios of the panels to set
     */
    public void setPanels(ArrayList<JPanel> panels, ArrayList<Double> panelRatios) {
        this.panels = panels;
        this.panelRatios = panelRatios;
        this.normalizeRatios(); // Normalize the ratios to be between 0 and 1... So any percentages become a ratio between 0 and 1 (50% = 0.5)
        this.dividers.clear();
        super.removeAll(); // Destroy the current panels and dividers and rebuild them with the new panels and their ratios

        // TODO: We do not need to remove everything, we can just get the list of panels and ratios and then rebuild the split pane with the new panels and ratios

        this.dividers = new ArrayList<>();

        for (int i = 0; i < panels.size() - 1; i++) {
            this.dividers.add(this.createDivider(this, i));
        }

        for (JPanel panel : panels) {
            super.add(panel);
        }

        for (JPanel divider : this.dividers) {
            super.add(divider);
        }

        this.resizeComponents();
        System.gc();
    }

    public void addResizeListener(CustomPanelResizeListener listener) {
        this.resizeListeners.add(listener);
    }

    public void removeResizeListener(CustomPanelResizeListener listener) {
        this.resizeListeners.remove(listener);
    }

    /**
     * Fires a resize event to all the resize listeners
     */
    protected void fireResizeEvent() {
        CustomPanelResizeEvent event = new CustomPanelResizeEvent(this);
        for (CustomPanelResizeListener listener : this.resizeListeners) {
            listener.customPanelResized(event);
        }
    }


    public ArrayList<Double> getPanelRatios() {
        return this.panelRatios;
    }

    public ArrayList<JPanel> getPanels() {
        return this.panels;
    }
}
