package edu.bu.signstream.common.util.https;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;
import edu.bu.signstream.common.util.https.requestBuilder.RequestBuilder;
import edu.bu.signstream.common.util.https.requestBuilder.wrappers.DataWrapper;
import edu.bu.signstream.common.util.https.requestBuilder.wrappers.FileWrapper;
import edu.bu.signstream.common.util.https.requestBuilder.wrappers.StringWrapper;
import edu.bu.signstream.common.util.https.responses.BasicResponse;
import edu.bu.signstream.ui.SS3SignStreamApplication;
import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.client.methods.*;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.client.LaxRedirectStrategy;
import org.apache.http.util.EntityUtils;

import java.io.IOException;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Predicate;
import java.util.logging.Level;


/**
 * This class is used to make basic HTTP requests
 */
public class HttpRequest {
    private String url; // The URL to send the request to

    private final CloseableHttpClient httpClient; // The HTTP client to use to send the request
    private final HttpUriRequest request; // The request to send
    private CloseableHttpResponse response; // The response from the request

    /**
     * @param request A Request object that specifies the type of request and the needed data for that request
     */
    public HttpRequest(RequestBuilder request) {
        this.url = request.getUrl();

        this.httpClient = HttpClients.custom().setRedirectStrategy(new LaxRedirectStrategy()).build();
        SS3SignStreamApplication.logger.log(Level.INFO, "Created default HTTP client");

        switch (request.getRequestType()) {
            case GET:
                this.request = new HttpGet(this.url);
                SS3SignStreamApplication.logger.log(Level.INFO, "Created HTTP GET request");
                break;
            case POST:
                this.request = new HttpPost(this.url);
                SS3SignStreamApplication.logger.log(Level.INFO, "Created HTTP POST request");
                break;
            case PUT:
                this.request = new HttpPut(this.url);
                SS3SignStreamApplication.logger.log(Level.INFO, "Created HTTP PUT request");
                break;
            case DELETE:
                this.request = new HttpDelete(this.url);
                SS3SignStreamApplication.logger.log(Level.INFO, "Created HTTP DELETE request");
                break;
            case PATCH:
                this.request = new HttpPatch(this.url);
                SS3SignStreamApplication.logger.log(Level.INFO, "Created HTTP PATCH request");
                break;
            default:
                throw new IllegalArgumentException("Request type must be GET, POST, PUT, DELETE or PATCH");
        }

        if (request.getFormLabel() != null && this.request instanceof HttpPost) { // If the request is a POST request, add the form label and the file to the request
            MultipartEntityBuilder builder = MultipartEntityBuilder.create(); // Create a builder for the POST request

            for (DataWrapper wrapper : request.getWrappers()) { // Add the file and form label to the builder
                if (wrapper instanceof FileWrapper) { // For post request, we currently support two types of wrappers, FileWrapper and StringWrapper, for files and strings respectively
                    FileWrapper fileWrapper = (FileWrapper) wrapper;
                    builder.addBinaryBody(fileWrapper.getFormLabel(), fileWrapper.getFile(), fileWrapper.getContentType(), fileWrapper.getFile().getName());
                } else if (wrapper instanceof StringWrapper) {
                    builder.addTextBody(wrapper.getFormLabel(), ((StringWrapper) wrapper).getTextBody(), wrapper.getContentType());
                }
            }

            ((HttpPost) this.request).setEntity(builder.build()); // Set the entity of the POST request to the builder
        }
    }

    /**
     * @return The BasicResponse of the POST request
     */
    public BasicResponse executeHttpRequest() throws IOException {
        this.response = this.httpClient.execute(this.request);
        return this.executeRequest();
    }


    /**
     * @param json The string we want to check if it is valid JSON
     * @return True if the string is valid JSON, false otherwise
     */
    public static boolean isJSONValid(final String json) {
        final ObjectMapper objectMapper = new ObjectMapper();
        boolean isValid = true;
        try {
            objectMapper.readTree(json);
        } catch (JsonProcessingException e) {
            isValid = false;
        }

        return isValid;
    }

    /**
     * This method should be used in the case that the JSON response from an endpoint is not valid.
     * @param jsonString The string we want to convert into an ugly JSON string
     * @return The ugly JSON string
     */
    public static String makeJSONUgly(String jsonString) {
        ObjectMapper mapper = new ObjectMapper();
        ObjectNode rootNode = mapper.createObjectNode();

        rootNode.put("data", jsonString.strip());

        String result = "";
        try {
            result = mapper.writerWithDefaultPrettyPrinter().writeValueAsString(rootNode);
        } catch (JsonProcessingException e) {
            SS3SignStreamApplication.logger.log(Level.SEVERE, "Error converting JSON string to ugly JSON string", e);
        }

        return result;
    }

    /**
     * Executes the request and returns a BasicResponse object
     * @return The BasicResponse of the request
     */
    private BasicResponse executeRequest() {
        BasicResponse basicResponse = null;

        try (CloseableHttpResponse response = this.response) {
            System.out.println(response.getStatusLine());
            HttpEntity responseEntity = response.getEntity();

            if (responseEntity != null) {
                System.out.println("Response content length: " + responseEntity.getContentLength());
                System.out.println("Response content type: " + responseEntity.getContentType());

                int statusCode = response.getStatusLine().getStatusCode();
                long contentLength = responseEntity.getContentLength();
                Header contentType = responseEntity.getContentType();
                String responseContent = EntityUtils.toString(responseEntity).strip();

                if (!isJSONValid(responseContent) && !responseContent.isEmpty() && !responseContent.isBlank()) { // If the response is not valid JSON, try to fix it
                    try {
                        /*
                           PLACE HOT FIXES TO MALFORMED JSON RESPONSES HERE
                           CODE HERE SHOULD NOT BE PERMANENT
                        */

                    } catch (Exception ignored) {
                        // This ugly function is only for debugging purposes... Don't use it for production
                        responseContent = makeJSONUgly(responseContent); // If we get a bad JSON response, we make it ugly to at least return some data back
                    }
                }

                System.out.println("Response content: " + responseContent);

                basicResponse = new BasicResponse(statusCode, contentLength, contentType, responseContent); // We create a basic response for any HTTP request, which is just a wrapper for the response along with the data represented as a string

                EntityUtils.consume(responseEntity); // This line may not be necessary, as you have consumed the responseEntity with EntityUtils.toString(responseEntity)
            }
        } catch (IOException e) {
            SS3SignStreamApplication.logger.log(Level.SEVERE, "Error executing request", e);
            throw new RuntimeException(e);
        }

        return basicResponse;
    }


    /**
     * Sets a timer for a producer and consumer thread to stop after a certain amount of time
     * @param producer The producer thread
     * @param consumer The consumer thread
     * @param threadLifeMilliseconds The amount of time the threads should run for before being interrupted and stopped
     */
    private static void scheduleExecutorService(Thread producer, Thread consumer, long threadLifeMilliseconds) {
        ScheduledExecutorService executor = Executors.newScheduledThreadPool(1); // Create a scheduled executor service with a single thread
        executor.schedule(() -> {
            if (producer.isAlive()) {
                producer.interrupt();
            }
            if (consumer.isAlive()) {
                consumer.interrupt();
            }
        }, threadLifeMilliseconds, TimeUnit.MILLISECONDS); // Schedule a task to interrupt the threads after a delay

        executor.shutdown();
    }

    /**
     * @param threadLifeMilliseconds The amount of time the threads should run for before being interrupted and stopped
     * @param pollingRateMilliseconds The amount of time the producer thread should wait before polling the endpoint again
     * @return The BasicResponse of the request, once a valid request is received or the producer thread times out due to length of time running determined by {@code threadLifeMilliseconds}
     */
    public BasicResponse waitForResponse(long threadLifeMilliseconds, long pollingRateMilliseconds, Predicate<BasicResponse> exitCondition) throws IOException {
        BlockingQueue<BasicResponse> queue = new LinkedBlockingQueue<>();
        AtomicReference<Exception> exceptionAtomicReference = new AtomicReference<>();
        AtomicReference<BasicResponse> finalResponse = new AtomicReference<>(); // We can probably make an anonymous inner class that implements runnable to get the value later, but this works. And it's probably fine.

        // Create a producer thread
        Thread producer = new Thread(() -> {
            while (true) {
                BasicResponse basicResponse;
                try {
                    basicResponse = this.executeHttpRequest();
                } catch (IOException e) {
                    exceptionAtomicReference.set(e);
                    break;
                }

                queue.add(basicResponse);

                try {
                    Thread.sleep(pollingRateMilliseconds);
                } catch (InterruptedException e) {
                    exceptionAtomicReference.set(e);
                    break;
                }
            }
        });

        Thread consumer = new Thread(() -> {
            try {
                while (true) {
                    // Check if exception is an IOException, if so, throw it
                    if (exceptionAtomicReference.get() instanceof IOException) {
                        break;
                    }

                    BasicResponse basicResponse = queue.take(); // This will block if queue is empty
                    finalResponse.set(basicResponse);
                    if (exitCondition.test(basicResponse)) {
                        producer.interrupt();
                        break;
                    }
                }
            } catch (InterruptedException e) {
                // Thread was interrupted because there's no more work to do
            }
        });

        producer.start();
        consumer.start();

        // Schedule a task to interrupt the threads after a delay
        scheduleExecutorService(producer, consumer, threadLifeMilliseconds);


        if (exceptionAtomicReference.get() instanceof IOException) {
            throw new IOException(exceptionAtomicReference.get());
        }

        try {
            producer.join(); // Wait for the threads to finish
            consumer.join();
        } catch (InterruptedException e) {
            SS3SignStreamApplication.logger.log(Level.SEVERE, "Error joining threads", e);
            producer.interrupt();
            consumer.interrupt();
        }

        return finalResponse.get();
    }

    /**
     * @param url Set URL to send the POST request to
     */
    public void setUrl(String url) {
        this.url = url;
    }
}
