package edu.bu.signstream.common.util.https.responses;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import edu.bu.signstream.common.util.Version;
import org.apache.http.Header;

/**
 * This class is used to store the response from a Choice API request. This is the JSON retrieved after the user has selected a sign on the DAI side.
 */
public class ChoiceResponse extends BasicResponse {
    private final int httpStatusCode;
    private final int choiceStatusCode;
    private final String message;
    private final ChoiceData data;
    private final String sessionID;
    private final String type;
    private final Version schemaVersion;
    private final Version dataVersion;
    private final String signText;
    private final ObjectMapper mapper;
    private final BaseChoiceResponse baseChoiceResponse;

    public ChoiceResponse(int statusCode, long contentLength, Header contentType, String contentJSON) throws JsonProcessingException {
        super(statusCode, contentLength, contentType, contentJSON);
        this.httpStatusCode = statusCode;

        this.mapper = new ObjectMapper(); // This is a Jackson object mapper. It is used to map JSON strings to Java objects.

        this.baseChoiceResponse = this.mapper.readValue(contentJSON, BaseChoiceResponse.class); // contentJSON is a JSON string represented as a normal string. Jackson will map this string to a BaseChoiceResponse object.
        this.choiceStatusCode = this.baseChoiceResponse.getChoiceResponseStatusCode();
        this.message = this.baseChoiceResponse.getMessage();
        this.data = this.baseChoiceResponse.getData(); // This is the ChoiceData object that contains the sessionID, type, schemaVersion, dataVersion, and signText. The contentJSON has a JSON inside of the surface level JSON, so we need to map the inner JSON to a ChoiceData object.
        this.sessionID = this.data.getSessionID();
        this.type = this.data.getType();
        this.schemaVersion = new Version(String.valueOf(this.data.getSchemaVersion()));
        this.dataVersion = new Version(String.valueOf(this.data.getDataVersion()));
        this.signText = this.data.getSignText();
    }

    public int getHttpStatusCode() {
        return this.httpStatusCode;
    }

    public int getChoiceStatusCode() {
        return this.choiceStatusCode;
    }

    public String getMessage() {
        return this.message;
    }

    public ChoiceData getData() {
        return this.data;
    }

    public String getSessionID() {
        return this.sessionID;
    }

    public String getType() {
        return this.type;
    }

    public Version getSchemaVersion() {
        return this.schemaVersion;
    }

    public Version getDataVersion() {
        return this.dataVersion;
    }

    public String getSignText() {
        return this.signText;
    }

    @Override
    public String toString() {
        try {
            return this.mapper.writerWithDefaultPrettyPrinter().writeValueAsString(this.baseChoiceResponse);
        } catch (JsonProcessingException e) {
            throw new RuntimeException(e);
        }
    }
}

/**
 * This is an intermediary class used to map the JSON response from the Choice API request to a Java object. The ChoiceResponse is just a wrapper around this class and other important objects exposed to the rest of the code. This code is only for the surface level JSON, not the other data in JSON string.
 * You can think of this class is step one, and the ChoiceResponse class as the holder and acquirer of all data from the Choice API request. It also gets data outside of this class.
 */
@JsonIgnoreProperties(ignoreUnknown = true)
class BaseChoiceResponse {
    @JsonProperty("status")
    private int choiceResponseStatusCode;

    @JsonProperty("message")
    private String message;

    @JsonProperty("data")
    private ChoiceData data;

    public int getChoiceResponseStatusCode() {
        return this.choiceResponseStatusCode;
    }

    public String getMessage() {
        return this.message;
    }

    public ChoiceData getData() {
        return this.data;
    }
}

/**
 * This is the inner JSON from the surface level JSON string. This is a part of the ChoiceResponse, a way to map and parse the inner JSON using the Jackson library.
 */
@JsonIgnoreProperties(ignoreUnknown = true)
class ChoiceData {
    @JsonProperty("session")
    private String sessionID;

    @JsonProperty("type")
    private String type;

    @JsonProperty("schema_version")
    private int schemaVersion;

    @JsonProperty("data_version")
    private int dataVersion;

    @JsonProperty("signText")
    private String signText;

    public String getSessionID() {
        return this.sessionID;
    }

    public String getType() {
        return this.type;
    }

    public int getSchemaVersion() {
        return this.schemaVersion;
    }

    public int getDataVersion() {
        return this.dataVersion;
    }

    public String getSignText() {
        return this.signText;
    }
}
