package edu.bu.signstream.grepresentation.view;

import edu.bu.signstream.common.util.CustomPanelizedSplitPane;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.Panel;

import java.awt.*;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;

public class ViewDataManager {
    private final ArrayList<ViewData> viewDataList;

    public ViewDataManager() {
        this.viewDataList = new ArrayList<>();
    }

    /**
     * Will sort the ArrayList of ViewData objects by order
     * @param viewDataList The ArrayList of ViewData objects
     */
    public ViewDataManager(ArrayList<ViewData> viewDataList) {
        viewDataList.sort(Comparator.comparingInt(ViewData::getOrder));
        this.viewDataList = viewDataList;
    }

    /**
     * @param viewData The ViewData object to add
     */
    public void addViewData(ViewData viewData) {
        this.viewDataList.add(viewData);
        this.viewDataList.sort(Comparator.comparingInt(ViewData::getOrder));
    }

    /**
     * This will add a new panel to the CustomPanelizedSplitPane. It will also take an equal amount of percentage from all the other panels
     * to add the new panel to the CustomPanelizedSplitPane in a proportional manner.
     * For example:
     * If there are two panels: panel1 and panel2 to an 60%:40% ratio, and we want to add a new panel, panel3, to the CustomPanelizedSplitPane,
     * then panel1 will be reduced to 43.33% and panel2 will be reduced to 23.33% to make room for panel3 to be 33.33% 1 / (2 + 1) = 0.3333
     * @param panel           The panel to add
     * @param outerPanelDimension The dimension of the CustomPanelizedSplitPane
     * @param panelOrder     The order of the panel to add
     */
    public void addPanel(Panel panel, Dimension outerPanelDimension, int panelOrder) {
        int minimumPanelHeight = CustomPanelizedSplitPane.MIN_PANEL_SIZE;
        double minimumRatio = (minimumPanelHeight / (double) outerPanelDimension.height) * 100.0; // The minimum height a panel is allowed to be
        double targetPanelPercentage = 100.0 / (this.viewDataList.size() + 1.0); // +1 for the new panel // How much percentage the new panel should take up based on how many existing panels there are
        double ratioAccumulator = 0.0;

        for (ViewData viewDataIterator : this.viewDataList) {
            double currentPanelPercentage = viewDataIterator.getPanelPercentage();
            double desiredQuantityToAdd = (100.0 / (this.viewDataList.size() + 1.0)) / this.viewDataList.size(); // How much percentage we want to take away from each existing panel to make room for the new panel

            if (currentPanelPercentage - desiredQuantityToAdd < minimumRatio) { // If taking this percentage away from the current panel will make it go below the minimum ratio, we just take as much as we can
                // In this case, we take as much as we can from the current panel without going below the minimum ratio
                double quantityToRemove = currentPanelPercentage - minimumRatio;
                viewDataIterator.setPanelPercentage(currentPanelPercentage - quantityToRemove);
                ratioAccumulator += quantityToRemove;
            } else {
                // In this case, we take the desired quantity from the current panel
                viewDataIterator.setPanelPercentage(currentPanelPercentage - desiredQuantityToAdd);
                ratioAccumulator += desiredQuantityToAdd;
            }

            if (ratioAccumulator >= targetPanelPercentage) {
                break;
            }
        }

        ViewData viewData = new ViewData(panel.id, panelOrder, ratioAccumulator); // The new panel will take up the amount of percentage that we took away from the other panels
        this.viewDataList.add(viewData);

        this.viewDataList.sort(Comparator.comparingInt(ViewData::getOrder));
    }

    /**
     * This will remove a panel from the CustomPanelizedSplitPane. It will also distribute the removed panel's percentage to the other panels evenly.
     * @param panel The panel to remove
     */
    public void removePanel(Panel panel) {
        double currentPanelPercentage = 0.0;
        for (ViewData viewDataIterator : this.viewDataList) {
            if (viewDataIterator.getPanelID().equals(panel.id)) {
                currentPanelPercentage = viewDataIterator.getPanelPercentage();
                this.viewDataList.remove(viewDataIterator);
                break;
            }
        }

        // Distribute the removed panel's percentage to the other panels evenly

        for (ViewData viewDataIterator : this.viewDataList) {
            double currentPercentage = viewDataIterator.getPanelPercentage();
            double newPercentage = currentPercentage + (currentPanelPercentage / this.viewDataList.size());
            viewDataIterator.setPanelPercentage(newPercentage);
        }
    }

    /**
     * @param viewData The ViewData object to remove
     */
    public void removeViewData(ViewData viewData) {
        this.viewDataList.remove(viewData);
    }

    /**
     * @param order The order to search for
     * @return The ViewData object with the specified order, or null if not found
     */
    public boolean removeViewData(int order) {
        for (ViewData viewDataIterator : this.viewDataList) {
            if (viewDataIterator.getOrder() == order) {
                this.viewDataList.remove(viewDataIterator);
                return true;
            }
        }

        return false;
    }

    /**
     * @param panelID The panel ID to remove
     * @return True if the ViewDataManager contains a ViewData object with the specified panel ID, false otherwise
     */
    public boolean removeViewData(String panelID) {
        for (ViewData viewDataIterator : this.viewDataList) {
            if (viewDataIterator.getPanelID().equals(panelID)) {
                this.viewDataList.remove(viewDataIterator);
                return true;
            }
        }

        return false;
    }

    /**
     * @param panelID The panel ID to search for
     * @return The ViewData object with the specified panel ID, or null if not found
     */
    public ViewData getViewData(String panelID) {
        for (ViewData viewDataIterator : this.viewDataList) {
            if (viewDataIterator.getPanelID().equals(panelID)) {
                return viewDataIterator;
            }
        }

        return null;
    }

    /**
     * @param panelID The panel ID to search for
     * @return True if the ViewDataManager contains a ViewData object with the specified panel ID, false otherwise
     */
    public boolean containsViewData(String panelID) {
        for (ViewData viewDataIterator : this.viewDataList) {
            if (viewDataIterator.getPanelID().equals(panelID)) {
                return true;
            }
        }

        return false;
    }

    /**
     * @return The ArrayList of ViewData objects
     */
    public ArrayList<ViewData> getViewDataList() {
        return this.viewDataList;
    }

    /**
     * @return The panel IDs as a Set
     */
    public Set<String> getPanelIDAsSet() {
        Set<String> panelIDSet = new HashSet<>();
        for (ViewData viewDataIterator : this.viewDataList) {
            panelIDSet.add(viewDataIterator.getPanelID());
        }
        return panelIDSet;
    }

    /*@Override
    public String toString() {
        StringBuilder output = new StringBuilder();
        for (ViewData viewDataIterator : this.viewDataList) {
            output.append(viewDataIterator.toString()).append("\n");
        }
        return output.toString();
    }*/
}
