package edu.bu.signstream.ui.panels.table.treetable.labels;

import javax.swing.tree.TreePath;

public class FileSystemModel extends AbstractTreeTableModel {

   // Names of the columns. size == row number in the table
   static protected String[] cNames = { "Name" }; 

   // Types of the columns. 
   static protected Class[] cTypes = { TreeTableModel.class }; 

   /** Node currently being reloaded, this becomes somewhat muddy if
    * reloading is happening in multiple threads. */
   protected FileNode reloadNode;


   /**
    * Creates a FileSystemModel2 with the root being <code>rootPath</code>.
    * This does not load it, you should invoke
    * <code>reloadChildren</code> with the root to start loading.
    */
   public FileSystemModel() {
      super(null);
      root = new FileNode("", 0);
      FileNode r = (FileNode)root;
      
      FileNode[] newChildren = new FileNode[3];
      newChildren[0] = new FileNode("node 4", 4);
      newChildren[1] = new FileNode("node 2", 2);
      newChildren[2] = new FileNode("node 3", 3);
      
      FileNode[] dNode1 = new FileNode[2];
      dNode1[0] = new FileNode("ghgh", 6);
      dNode1[1] = new FileNode("ghgh", 7);
      
      newChildren[2].setChildren(dNode1, false);
      
      r.setChildren(newChildren, true);
   }

   //
   // The TreeModel interface
   //

   /**
    * Returns the number of children of <code>node</code>.
    */
   @Override
public int getChildCount(Object node) {
      Object[] children = getChildren(node);
      return (children == null) ? 0 : children.length;
   }

   /**
    * Returns the child of <code>node</code> at index <code>i</code>.
    */
   @Override
public Object getChild(Object node, int i) {
      return getChildren(node)[i];
   }

   /**
    * Returns true if the passed in object represents a leaf, false
    * otherwise.
    */
   @Override
public boolean isLeaf(Object node) {
      return ((FileNode) node).isLeaf();
   }

   //
   //  The TreeTableNode interface. 
   //

   /**
    * Returns the number of columns.
    */
   @Override
public int getColumnCount() {
      return cNames.length;
   }

   /**
    * Returns the name for a particular column.
    */
   @Override
public String getColumnName(int column) {
      return cNames[column];
   }

   /**
    * Returns the class for the particular column.
    */
   @Override
public Class getColumnClass(int column) {
      return cTypes[column];
   }

   /**
    * Returns the value of the particular column.
    */
   @Override
public Object getValueAt(Object node, int column) {
      FileNode fn = (FileNode) node;

      try {
         switch (column) {
            case 0 :
               return fn.getLabel();
         }
      } catch (SecurityException se) {}

      return null;
   }

   /**
    * Returns the path to the node that is being loaded.
    */
   public TreePath getPathLoading() {
      FileNode rn = reloadNode;

      if (rn != null) {
         return new TreePath(rn.getPath());
      }
      return null;
   }

   /**
    * Returns the node being loaded.
    */
   public Object getNodeLoading() {
      return reloadNode;
   }

   protected Object[] getChildren(Object node) {
      FileNode fileNode = ((FileNode) node);
      return fileNode.getChildren();
   }

   protected static FileNode[] EMPTY_CHILDREN = new FileNode[0];

   /**
    * A FileNode is a derivative of the File class - though we delegate to
    * the File object rather than subclassing it. It is used to maintain a
    * cache of a directory's children and therefore avoid repeated access
    * to the underlying file system during rendering. 
    */
   class FileNode {
      /** Parent FileNode of the receiver. */
      private FileNode parent;
      /** Children of the receiver. */
      protected FileNode[] children;
      
      protected String label = "";
      protected int rowNumber = 0;

      protected FileNode(String label, int rowNumber) {
         this(null, label, rowNumber);
      }

      protected FileNode(FileNode parent, String label, int rowNumber) {
         this.parent = parent;
         this.label = label;
         this.rowNumber = rowNumber;        
      }
      
      public String getLabel() {
         return label;
      }
      
      public int getRowNumber() {
         return rowNumber;
      }

      /**
       * Returns the the string to be used to display this leaf in the JTree.
       */
      @Override
	public String toString() {
         return label;
      }

      /**
       * Returns the parent of the receiver.
       */
      public FileNode getParent() {
         return parent;
      }

      /**
       * Returns true if the receiver represents a leaf, that is it is
       * isn't a directory.
       */
      public boolean isLeaf() {
         return children == null || children.length == 0; 
      }

      /**
       * Loads the children, caching the results in the children
       * instance variable.
       */
      protected FileNode[] getChildren() {
         return children;
      }

      /**
       * Gets the path from the root to the receiver.
       */
      public FileNode[] getPath() {
         return getPathToRoot(this, 0);
      }

      protected FileNode[] getPathToRoot(FileNode aNode, int depth) {
         FileNode[] retNodes;

         if (aNode == null) {
            if (depth == 0)
               return null;
            else
               retNodes = new FileNode[depth];
         } else {
            depth++;
            retNodes = getPathToRoot(aNode.getParent(), depth);
            retNodes[retNodes.length - depth] = aNode;
         }
         return retNodes;
      }

      /**
       * Sets the children of the receiver, updates the total size,
       * and if generateEvent is true a tree structure changed event
       * is created.
       */
      protected void setChildren(
         FileNode[] newChildren,
         boolean generateEvent) {

         children = newChildren;

         if (generateEvent) {
            FileNode[] path = getPath();
            fireTreeStructureChanged(FileSystemModel.this, path, null, null);
         }
      }

      /**
       * Can be invoked when a node has changed, will create the
       * appropriate event.
       */
      protected void nodeChanged() {
         FileNode parent = getParent();

         if (parent != null) {
            FileNode[] path = parent.getPath();
            int[] index = { getIndexOfChild(parent, this)};
            Object[] children = { this };

            fireTreeNodesChanged(FileSystemModel.this, path, index, children);
         }
      }
   }
}
